#!/usr/bin/env python
# encoding: utf-8
"""
*This is a template gocart plugin script. Duplicate this script, give it a name starting with 'gp_' (e.g. 'gp_slack_alert.py' and modifiy the code to do something magical everytime a GCN alert is read.*

:Author:
    David Young

:Date Created:
    May 11, 2023

Usage:
    gp_template <alertDir>

Options:
    alertDir              path to an alert directory

    -h, --help            show this help message
    -v, --version         show version
"""
################# GLOBAL IMPORTS ####################
import sys
import os
from fundamentals import tools


def plugin(
        log,
        settings,
        alertFiles,
        alertMeta):
    """*this is the gocart plugin function that will be run when an alert is read*

    **Key Arguments:**

    - ``log`` -- logger
    - ``settings`` -- these are the gocart settings, you can add extra settings to the gocart.yaml settings file and they will be read here.
    - ``alertFiles`` -- a list of all the files generated by gocart for the alert
    - ``alertMeta`` -- a dictionary of the alert metadata from the json alert, FITS Header and gocart generated extras.      
    """
    log.debug('starting the ``plugin`` function')

    alertType = alertMeta["ALERT"]["alert_type"]
    evertId = alertMeta["ALERT"]["superevent_id"]

    print(f"Here are the alert files generated from {alertType} alert of event {evertId}:")
    for f in alertFiles:
        print(f)

    log.debug('completed the ``plugin`` function')
    return None


# DO NOT EDIT ANYTHING BOTH THIS LINE
def main(arguments=None):
    """
    *The main function used when ``gp_template.py`` is run as a single script from the cl*
    """

    # SETUP THE COMMAND-LINE UTIL SETTINGS
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="WARNING",
        options_first=False,
        projectName="gocart",
        defaultSettingsFile=True
    )
    arguments, settings, log, dbConn = su.setup()

    # UNPACK REMAINING CL ARGUMENTS USING `EXEC` TO SETUP THE VARIABLE NAMES
    # AUTOMATICALLY
    a = {}
    for arg, val in list(arguments.items()):
        if arg[0] == "-":
            varname = arg.replace("-", "") + "Flag"
        else:
            varname = arg.replace("<", "").replace(">", "")
        a[varname] = val
        if arg == "--dbConn":
            dbConn = val
            a["dbConn"] = val
        log.debug('%s = %s' % (varname, val,))

    # GENERATE A LIST OF FILE PATHS
    pathToDirectory = a["alertDir"]
    alertFiles = []
    alertMeta = None
    for d in os.listdir(pathToDirectory):
        filepath = os.path.join(pathToDirectory, d)
        if os.path.isfile(filepath) and d[0] != ".":
            alertFiles.append(d)

        if d == "meta.yaml":
            import yaml
            # ADD YAML CONTENT TO DICTIONARY
            with open(filepath, 'r') as stream:
                alertMeta = yaml.safe_load(stream)

    plugin(
        log=log,
        settings=settings,
        alertFiles=alertFiles,
        alertMeta=alertMeta
    )

    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# encoding: utf-8
"""
Documentation for gocart can be found here: http://gocart.readthedocs.org

Usage:
    gocart init
    gocart [-p] echo <daysAgo> [-s <pathToSettingsFile>]
    gocart [-p] (listen|quit|restart|status) [-s <pathToSettingsFile>]


Options:
    init                                   setup the gocart settings file for the first time
    echo <daysAgo>                         relisten to alerts from N <daysAgo> until now and then exit
    listen                                 reconnect to kafka stream and listen from where you left off (or from now on if connectiong for the first time).

    -h, --help                             show this help message
    -v, --version                          show version
    -p, --plugins                          execute plugins everytime an alert is read
    -s, --settings <pathToSettingsFile>    the settings file
    -t, --test                             test, only collect 1 map
"""
from subprocess import Popen, PIPE, STDOUT
from fundamentals import tools, times
from docopt import docopt
import pickle
import glob
import readline
import sys
import os
import shutil
os.environ['TERM'] = 'vt100'


def tab_complete(text, state):
    return (glob.glob(text + '*') + [None])[state]


def main(arguments=None):
    """
    *The main function used when `cl_utils.py` is run as a single script from the cl, or when installed as a cl command*
    """
    # setup the command-line util settings
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="WARNING",
        options_first=False,
        projectName="gocart",
        defaultSettingsFile=True
    )
    arguments, settings, log, dbConn = su.setup()

    # tab completion for raw_input
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind("tab: complete")
    readline.set_completer(tab_complete)

    # UNPACK REMAINING CL ARGUMENTS USING `EXEC` TO SETUP THE VARIABLE NAMES
    # AUTOMATICALLY
    a = {}
    for arg, val in list(arguments.items()):
        if arg[0] == "-":
            varname = arg.replace("-", "") + "Flag"
        else:
            varname = arg.replace("<", "").replace(">", "")
        a[varname] = val
        log.debug('%s = %s' % (varname, val,))

    firstConnect = False
    if "gcn-kafka" in settings and (not settings["gcn-kafka"]["group_id"] or len(str(settings["gcn-kafka"]["group_id"])) < 7):

        from os.path import expanduser
        import uuid as pyuuid
        import re
        group_id = pyuuid.uuid1().int
        settings["gcn-kafka"]["group_id"] = group_id

        home = expanduser("~")
        filepath = home + "/.config/gocart/gocart.yaml"
        import codecs
        with codecs.open(filepath, encoding='utf-8', mode='r') as readFile:
            content = readFile.read()
            regex = re.compile(r'group_id\:.*')
            content = regex.sub(f"group_id: {group_id}", content, count=1)
        with codecs.open(filepath, encoding='utf-8', mode='w') as writeFile:
            writeFile.write(content)
        firstConnect = True
    elif "gcn-kafka" not in settings:
        return

    ## START LOGGING ##
    startTime = times.get_now_sql_datetime()
    log.info(
        '--- STARTING TO RUN THE cl_utils.py AT %s' %
        (startTime,))

    # TEST IF PLUG FOLDER EXISTS
    from os.path import expanduser
    home = expanduser("~")
    pluginPath = home + "/.config/gocart/plugins"
    exists = os.path.exists(pluginPath)
    if not exists:
        shutil.copytree(os.path.dirname(__file__) + "/resources/plugins", pluginPath)

    if a["init"]:
        from os.path import expanduser
        home = expanduser("~")
        filepath = home + "/.config/gocart/gocart.yaml"

    topic = 'igwn.gwalert'

    if len(settings['gcn-kafka']['client_id']) < 6 or len(settings['gcn-kafka']['client_secret']) < 6:
        print("Please add your gcn-kafka client ID and secret to the gocart.yaml settings file.")
        return

    if a['listen'] or a['echo']:

        if a['listen']:
            verb = "Listening for"
        else:
            verb = "Echoing"

        parse_mock_events = settings["lvk"]["parse_mock_events"]
        parse_real_events = settings["lvk"]["parse_real_events"]
        # WHERE TO DOWNLOAD MAPS TO
        if "download_dir" in settings["lvk"] and settings["lvk"]["download_dir"]:
            download_dir = settings["lvk"]["download_dir"]
            # MAKE RELATIVE HOME PATH ABSOLUTE
            from os.path import expanduser
            home = expanduser("~")
            if download_dir == "~":
                download_dir = download_dir.replace("~", home)
        else:
            download_dir = "."

        mockDir = download_dir + "/mockevents/"
        evertDir = download_dir + "/superevents/"
        if parse_mock_events:
            print(f"{verb} Mockevents")
            if not os.path.exists(mockDir):
                os.makedirs(mockDir)
        if parse_real_events:
            print(f"{verb} Superevents")
            if not os.path.exists(evertDir):
                os.makedirs(evertDir)

    if a['listen'] or a["quit"] or a["status"] or a["restart"]:

        # ADD SOMETHING LIKE THE FOLLOWING TO THE CL USAGE:
        """
        Usage:
            myCommand (start|stop|restart|status)

            Options:
                start                 start the myCommand daemon
                stop                  stop the myCommand daemon
                restart               restart the myCommand daemon
                status                print the staus of the myCommand daemon
        """

        from fundamentals import daemonise

        class myDaemon(daemonise):
            def action(
                    self,
                    **kwargs):
                self.log.info('starting the ``action`` method')

                from gcn_kafka import Consumer
                from confluent_kafka import TopicPartition
                from gocart.parsers import lvk

                firstConnect = kwargs["firstConnect"]

                from datetime import datetime, date, time
                now = datetime.now()
                now = now.strftime("%Y%m%dt%H%M%S")

                print(f"gocart listen started at {now}")

                config = {
                    'group.id': settings["gcn-kafka"]["group_id"],
                    'enable.auto.commit': False,
                    'auto.offset.reset': 'earliest'
                }

                consumer = Consumer(config=config, client_id=settings['gcn-kafka']['client_id'],
                                    client_secret=settings['gcn-kafka']['client_secret'], domain='gcn.nasa.gov')
                consumer.subscribe([topic])

                stop = False
                test = 0
                more = True
                while not stop:
                    # IF FISRT TIME CONNECTING THEN SKIP MESSAGES
                    if firstConnect:
                        count = 0
                        index = 1
                        print("Marking previous messages as read, this can take a few minutes")
                        while more:
                            messages = consumer.consume(num_messages=300, timeout=5)
                            for message in messages:
                                count += 1
                                consumer.commit(message)
                            if index > 1:
                                # Cursor up one line and clear line
                                sys.stdout.write("\x1b[1A\x1b[2K")
                            index += 1
                            print(f"{count} messages read ...")
                            if not len(messages):
                                more = False

                        firstConnect = False
                        print(f"This is your first time using the listen command. gocart will now listen for all new incoming alerts (skipping the {count} previous alerts currently in this topic). If you stop listening and restart sometime later, gocart will immediately collect all alerts missed while off-line.")
                    for message in consumer.consume(timeout=5):
                        parser = lvk(
                            log=log,
                            record=message.value(),
                            settings=settings,
                            plugins=a["pluginsFlag"]
                        ).parse()
                        consumer.commit(message)

                self.log.info('completed the ``action`` method')
                return None

        d = myDaemon(log=log, name="gocart", firstConnect=firstConnect)

        if a['listen']:
            d.start()
        elif a['quit']:
            d.stop()
        elif a['restart']:
            d.restart()
        elif a['status']:
            d.status()

    if a['echo'] and a['daysAgo']:
        # GET MESSAGES OCCURRING IN LAST N DAYS
        from gcn_kafka import Consumer
        from confluent_kafka import TopicPartition
        from gocart.parsers import lvk
        import datetime

        consumer = Consumer(client_id=settings['gcn-kafka']['client_id'],
                            client_secret=settings['gcn-kafka']['client_secret'], domain='gcn.nasa.gov')

        since_utc = datetime.datetime.now() - datetime.timedelta(days=float(a['daysAgo']))
        timestamp1 = int((since_utc).timestamp() * 1000)
        since_utc = since_utc.strftime("%Y-%m-%d %H:%M:%S")
        print(f"Echoing alerts since {since_utc} UTC")

        start = consumer.offsets_for_times(
            [TopicPartition(topic, 0, timestamp1)])

        consumer.assign(start)

        more = True
        while more:
            messages = consumer.consume(num_messages=1, timeout=10)
            for message in messages:
                parser = lvk(
                    log=log,
                    record=message.value(),
                    settings=settings,
                    plugins=a["pluginsFlag"]
                ).parse()
            if not len(messages):
                more = False

    ## FINISH LOGGING ##
    endTime = times.get_now_sql_datetime()
    runningTime = times.calculate_time_difference(startTime, endTime)
    log.info('-- FINISHED ATTEMPT TO RUN THE cl_utils.py AT %s (RUNTIME: %s) --' %
             (endTime, runningTime, ))

    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# encoding: utf-8
"""
*Convert Healpix Map to ascii map format*

:Author:
    David Young

:Date Created:
    March 29, 2023
"""
from gocart.commonutils import flatten_healpix_map
from fundamentals import tools
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'


class ascii(object):
    """
    *Take a healpix map and convert to an ascii map with one row per ~deg2*

    **Key Arguments:**
        - ``log`` -- logger
        - ``mapPath`` -- path the the healpix map or an astropy skymap table
        - ``nside`` -- size of healpix pixels to resolve the sky to
        - ``settings`` -- the settings dictionary

    **Usage:**

    To setup your logger and settings, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    To convert a healpix map to ascii, run:

    ```python
    from gocart.convert import ascii
    c = ascii(
        log=log,
        mapPath="/path/to/bayestar.multiorder.fits",
        nside=64,
        settings=settings
    )
    asciiContent = c.convert(outputFilepath="/path/to/skymap.csv")
    ```
    """

    def __init__(
            self,
            log,
            mapPath,
            nside=64,
            settings=False,

    ):
        self.log = log
        log.debug("instansiating a new 'ascii' object")
        self.settings = settings
        self.mapPath = mapPath
        self.nside = nside

        self.table = flatten_healpix_map(
            log=log,
            mapPath=self.mapPath,
            nside=self.nside
        )

        return None

    def convert(
            self,
            outputFilepath=False):
        """
        *Convert the healpix map to ascii format and optionally save the ascii map to file*

        **Key Arguments:**
            - ``outputFilepath`` -- optionally write content to file. Default *False*

        **Return:**
            - ``ascii`` -- the CSV version of the healpix file (nside=64)
        """
        self.log.debug('starting the ``get`` method')

        import astropy_healpix as ah
        from astropy.coordinates import SkyCoord

        tableData = self.table

        # CREATE RA AND DEC COLUMNS
        ra, dec = ah.healpix_to_lonlat(tableData.index, 64, order='nested')
        tableData["RA"] = ra.deg
        tableData["DEC"] = dec.deg

        # ALSO GLON GLAT COLUMNS
        galacticCoords = SkyCoord(ra.deg, dec.deg, frame='icrs', unit='deg').galactic
        tableData["GLON"] = galacticCoords.l.degree
        tableData["GLAT"] = galacticCoords.b.degree

        # DROP COLUMNS
        try:
            tableData.drop(columns=['DISTNORM'], inplace=True)
        except:
            pass

        # CREATE HEADER FOR FILE
        try:
            header = f"# EVENT:{self.table.meta['objid']}\n"
        except:
            eventId = self.mapPath.split("/")[-3]
            header = f"# EVENT:{eventId}\n"
        header += f"# NSIDE:{self.nside}\n"

        tableData.index.names = ['IPIX']
        asciiContent = tableData.to_csv(index=True)
        asciiContent = header + asciiContent

        if outputFilepath:
            import codecs
            with codecs.open(outputFilepath, encoding='utf-8', mode='w') as writeFile:
                writeFile.write(asciiContent)

        self.log.debug('completed the ``get`` method')
        return asciiContent



#!/usr/bin/env python
# encoding: utf-8
"""
*Generate some extra stats for a given Healpix map*

:Author:
    David Young

:Date Created:
    March 30, 2023
"""
from fundamentals import tools
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'


def generate_skymap_stats(
        skymap,
        log):
    """*Generate some extra stats for a given Healpix map*

    **Key Arguments:**
        - ``skymap`` -- either a path to a healpix map FITS file or a skymap in astropy table format
        - ``log`` -- logger

    **Return:**
        - ``extras`` -- a diction of value added map stats

    ```python
    from gocart.commonutils import generate_skymap_stats
    extras = generate_skymap_stats(
        log=log,
        skymap="path/to/bayestar.multiorder.fits",
    )
    ```
    """
    log.debug('starting the ``generate_skymap_stats`` function')

    from astropy.table import Table
    import astropy_healpix as ah
    import astropy.units as u
    import numpy as np
    if isinstance(skymap, str):
        skymap = Table.read(skymap)

    tableData = skymap.to_pandas()

    # FIND LEVEL AND NSIDE PIXEL INDEX FOR EACH MULTI-RES PIXEL
    tableData['LEVEL'], tableData['IPIX'] = ah.uniq_to_level_ipix(tableData['UNIQ'])
    tableData['NSIDE'] = ah.level_to_nside(tableData['LEVEL'])
    # DETERMINE THE PIXEL AREA AND PROB OF EACH PIXEL
    tableData['AREA'] = ah.nside_to_pixel_area(tableData['NSIDE']).to_value(u.steradian)
    tableData['PROB'] = tableData['AREA'] * tableData["PROBDENSITY"]
    tableData['AREA'] = ah.nside_to_pixel_area(tableData['NSIDE']).to_value(u.deg**2)

    # SORT BY PROB, CALCULATE CUMULATIVE PROB AND RESORT BY INDEX
    tableData.sort_values(["PROBDENSITY"],
                          ascending=[False], inplace=True)
    tableData["CUMPROB"] = np.cumsum(tableData['PROB'])

    # FILTER DATA FRAME
    # FIRST CREATE THE MASK
    mask = (tableData["CUMPROB"] < 0.9)
    area90 = tableData.loc[mask, 'AREA'].sum()
    mask = (tableData["CUMPROB"] < 0.5)
    area50 = tableData.loc[mask, 'AREA'].sum()
    mask = (tableData["CUMPROB"] < 0.1)
    area10 = tableData.loc[mask, 'AREA'].sum()
    mask = (tableData["CUMPROB"] < 0.99999999999999)
    area100 = tableData['AREA'].sum()

    area90 = float(f'{area90:.3f}')
    area50 = float(f'{area50:.3f}')
    area10 = float(f'{area10:.3f}')

    extras = {"area90": area90, "area50": area50, "area10": area10}

    level, ipix = ah.uniq_to_level_ipix(
        skymap[np.argmax(skymap['PROBDENSITY'])]['UNIQ']
    )
    ra, dec = ah.healpix_to_lonlat(ipix, ah.level_to_nside(level),
                                   order='nested')

    from astropy.coordinates import SkyCoord
    galacticCoords = SkyCoord(ra, dec, frame='icrs').galactic
    glon = galacticCoords.l.degree
    glat = galacticCoords.b.degree

    extras["central coordinate"] = {
        "equatorial": f"{ra.deg:.6f} {dec.deg:.6f}",
        "galactic": f"{glon:.6f} {glat:.6f}"
    }

    log.debug('completed the ``generate_skymap_stats`` function')
    return extras



#!/usr/bin/env python
# encoding: utf-8
"""
*Plot multiorder FITS in an aitoff projection*

:Author:
    David Young

:Date Created:
    March 21, 2023
"""
from fundamentals import tools
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'


class aitoff(object):
    """
    *Convert Healpix Map to Aitoff projection with options to include galactic plane and sun position*

    **Key Arguments:**
        - ``log`` -- logger
        - ``mapPath`` -- path to the multiorder FITS file
        - ``outputFolder`` -- path to output the results to
        - ``settings`` -- the settings dictionary
        - ``meta`` -- extra meta data to present on plots. Default: {}

    **Usage:**

    To setup your logger and settings, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    To print the healpix map as an aitoff projection, use the following:

    ```python
    from gocart.convert import aitoff
    converter = aitoff(
        log=log,
        mapPath="path/to/bayestar.multiorder.fits",
        settings=settings
    )
    converter.convert()
    ```
    """

    def __init__(
            self,
            log,
            mapPath,
            outputFolder,
            settings=False,
            meta={}
    ):
        self.log = log
        log.debug("instansiating a new 'aitoff' object")
        self.settings = settings
        self.mapPath = mapPath
        self.outputFolder = outputFolder
        self.meta = meta
        # xt-self-arg-tmpx

        return None

    def convert(
        self,
        contours=True,
        galacticPlane=True,
        sunmoon=True,
        sunmoonContour=True
    ):
        """
        *convert the healpix map to an aitoff plot*

        **Key Arguments:**
            - ``contours`` -- plot 50 and 90% contours. Default *True*
            - ``galacticPlane`` -- plot galactic plane contours. Default *True*
            - ``sunmoon`` -- plot sun and moon. Default *True*
            - ``sunmoonContour`` -- show contours within 33 deg of sun and 20 deg from moon

        **Return:**
            - ``plotPath`` -- path to the printed plot
        """
        self.log.debug('starting the ``convert`` method')

        import matplotlib.pyplot as plt
        from . import healpix2cart
        import astropy.units as u
        import numpy as np
        from astropy.coordinates import SkyCoord, Galactic, get_sun, get_moon
        from matplotlib.projections.geo import GeoAxes
        from matplotlib.cm import get_cmap
        import matplotlib.patches as mpatches
        from matplotlib.lines import Line2D
        from astropy.time import Time

        import matplotlib
        matplotlib.use('PDF')

        class ThetaFormatterShiftPi(GeoAxes.ThetaFormatter):
            """SHIFTS LABELLING BY PI
            SHIFTS LABELLING FROM -180,180 TO 360-0"""

            def __call__(self, x, pos=None):
                x -= np.pi
                x = -x
                return GeoAxes.ThetaFormatter.__call__(self, x, pos)

        # CONVERT TO RECTILINEAR GRID
        converter = healpix2cart(
            log=self.log,
            mapPath=self.mapPath,
            settings=self.settings
        )
        wcs, mapDF, header = converter.convert()

        xsize = mapDF["PIXEL_X"].max() - mapDF["PIXEL_X"].min() + 1
        ysize = mapDF["PIXEL_Y"].max() - mapDF["PIXEL_Y"].min() + 1

        # RA RANGES FROM 0-360 ... NEED TO FLIP 360-0, AND THEN SHIFT BY 180 TO MATCH MATPLOTLIB FRAME
        mapDF = mapDF.iloc[::-1].reset_index()
        mapDF["RASHIFTED"] = -mapDF["RA"] + 180
        data = mapDF["PROB"].values.reshape((ysize, xsize))
        long = np.deg2rad(mapDF["RASHIFTED"].values).reshape((ysize, xsize))
        lat = np.deg2rad(mapDF["DEC"].values).reshape((ysize, xsize))

        # long = mapDF["RASHIFTED"].values.reshape((ysize, xsize))
        # lat = mapDF["DEC"].values.reshape((ysize, xsize))

        # MATPLOTLIB IS DOING THE PROJECTION
        cmap = get_cmap("hot_r")
        fig = plt.figure()

        # AITOFF DOES NOT PLAY WELL WITH ADDING LABELS - USE HAMMER PROJECTION INSTEAD
        ax = fig.add_subplot(111, projection='hammer')

        # RASTERIZED MAKES THE MAP BITMAP WHILE THE LABELS REMAIN VECTORIAL
        std = data.std()
        mean = data.mean()
        # image = ax.pcolormesh(long, lat, data, rasterized=False, cmap=cmap, vmin=mean, vmax=mean + 5 * std)

        # GRATICULE
        ax.set_longitude_grid(30)
        ax.set_latitude_grid(15)
        ax.xaxis.set_major_formatter(ThetaFormatterShiftPi(30))
        ax.set_longitude_grid_ends(90)

        if sunmoonContour:
            t = Time(header['DATE-OBS'], scale='utc')

            # FIND SUN AND PLACE ON CORRECT PLOT COORDINATE
            sun = get_sun(t)
            sun.ra.degree = -sun.ra.degree + 180
            if sun.ra.degree > 180.:
                sun.ra.degree -= 360
            sun.ra.radian = np.deg2rad(sun.ra.degree)

            # COMPUTE LONS AND LATS OF DAY/NIGHT TERMINATOR.
            nlons = 1441
            nlats = ((nlons - 1) / 2) + 1

            # COLOR IN THE SUN
            lons2 = np.linspace(-np.pi, np.pi, nlons)
            lats2 = np.linspace(-np.pi / 2, np.pi / 2, int(nlats))
            lons2, lats2 = np.meshgrid(lons2, lats2)
            sunlight = np.ones(lons2.shape)
            raSep = sun.ra.radian - lons2
            raSep[raSep > np.pi] = 2 * np.pi - raSep[raSep > np.pi]
            separation = ((raSep * np.cos((sun.dec.radian + lats2) / 2))**2 + (sun.dec.radian - lats2)**2)**0.5
            sunlight[separation < np.radians(33)] = 0
            sunyellow = matplotlib.colors.colorConverter.to_rgba('#b58900', alpha=0.2)
            ax.contourf(lons2, lats2, sunlight, 1, colors=[sunyellow, (0.0, 0.0, 0.0, 0.0)], zorder=3)

            # PLOT THE MOON
            moon = get_moon(t)
            moon.ra.degree = -moon.ra.degree + 180
            if moon.ra.degree > 180.:
                moon.ra.degree -= 360
            moon.ra.radian = np.deg2rad(moon.ra.degree)

            # COMPUTE LONS AND LATS OF DAY/NIGHT TERMINATOR.
            nlons = 1441
            nlats = ((nlons - 1) / 2) + 1

            # COLOR IN THE SUN
            lons2 = np.linspace(-np.pi, np.pi, nlons)
            lats2 = np.linspace(-np.pi / 2, np.pi / 2, int(nlats))
            lons2, lats2 = np.meshgrid(lons2, lats2)
            moonlight = np.ones(lons2.shape)
            raSep = moon.ra.radian - lons2
            raSep[raSep > np.pi] = 2 * np.pi - raSep[raSep > np.pi]
            separation = ((raSep * np.cos((moon.dec.radian + lats2) / 2))**2 + (moon.dec.radian - lats2)**2)**0.5
            moonlight[separation < np.radians(20)] = 0
            moonblue = matplotlib.colors.colorConverter.to_rgba('#268bd2', alpha=0.2)
            ax.contourf(lons2, lats2, moonlight, 1, colors=[moonblue, (0.0, 0.0, 0.0, 0.0)], zorder=3)

        # PLOT THE SUN
        if sunmoon:
            t = Time(header['DATE-OBS'], scale='utc')

            # FIND SUN AND PLACE ON CORRECT PLOT COORDINATE
            sun = get_sun(t)
            sun.ra.degree = -sun.ra.degree + 180
            if sun.ra.degree > 180.:
                sun.ra.degree -= 360
            sun.ra.radian = np.deg2rad(sun.ra.degree)

            label = "Sun"
            if sunmoonContour:
                label += " (within $33^o$)"
            ax.scatter(sun.ra.radian, sun.dec.radian, color="#b58900", alpha=0.8, s=20, marker="o", edgecolors="#cb4b16", linewidths=0.5, label=label, zorder=30)

            # PLOT THE MOON
            moon = get_moon(t)
            moon.ra.degree = -moon.ra.degree + 180
            if moon.ra.degree > 180.:
                moon.ra.degree -= 360
            moon.ra.radian = np.deg2rad(moon.ra.degree)

            label = "Moon"
            if sunmoonContour:
                label += " (within $20^o$)"
            ax.scatter(moon.ra.radian, moon.dec.radian, color="#268bd2", alpha=0.8, s=20, marker="o", edgecolors="#1e6ea7", linewidths=0.5, label=label, zorder=29)

        handles, labels = plt.gca().get_legend_handles_labels()
        if galacticPlane:
            # LON:LAT is lon 0-360 at lat=0
            lon_array = np.arange(0, 360, 0.5)
            lat_arry = np.full_like(lon_array, 0)
            galc = SkyCoord(l=lon_array, b=lat_arry, frame=Galactic, unit=u.deg)

            # CONVERT TO EQUATORIAL COORDINATES
            equatorial_array = galc.icrs
            gRa = equatorial_array.ra.degree
            gDec = equatorial_array.dec.degree
            gx = np.remainder(gRa + 180, 360)
            # SCALE CONVERSION TO [-180, 180] & FLIP RA
            ind = gx > 180
            gx[ind] -= 360
            gx = -gx
            planeColour = "#dc322f"
            ax.scatter(np.radians(gx), np.radians(gDec),
                       color=planeColour, alpha=1, s=1)
            line = Line2D([0], [0], label='Galactic Plane', color=planeColour)
            handles.append(line)

        if contours:
            mapDF.sort_values(["PROBDENSITY"],
                              ascending=[False], inplace=True)
            mapDF["CUMPROB"] = np.cumsum(mapDF['PROB'])
            mapDF["CUMPROB"] = mapDF["CUMPROB"] * 100. * mapDF["PROB"].sum()
            mapDF.sort_index(inplace=True)
            contours = mapDF["CUMPROB"].values.reshape((ysize, xsize))

            colors = ['#474973', '#a69cac', '#03F7EB']
            levels = [90, 50, 10]

            z = 10

            for c, l in zip(colors, levels):
                z += 1
                ax.contourf(long, lat,
                            contours, levels=[0, l], colors=c, zorder=z, alpha=0.8)
                if "EXTRA" in self.meta and f"area{l}" in self.meta["EXTRA"]:
                    area = self.meta["EXTRA"][f"area{l}"]
                    label = f"{l}%: {area:.1f} deg$^2$"
                else:
                    label = f"{l}%"
                patch = mpatches.Patch(color=c, label=label)
                handles.append(patch)

        if len(self.meta):
            data = ""
            data += f"Event: {self.meta['ALERT']['superevent_id']} ({self.meta['ALERT']['event']['group']})\n"
            eventDate = Time(self.meta['HEADER']['DATE-OBS'], scale='utc').to_datetime().strftime("%Y-%m-%d %H:%M:%S")
            data += f"Event time: {eventDate}\n"

            data += "\n"
            data += f"Alert: {self.meta['ALERT']['alert_type'].replace('_',' ')}\n"
            alertDate = Time(self.meta['ALERT']['time_created'], scale='utc').to_datetime().strftime("%Y-%m-%d %H:%M:%S")
            data += f"Alert time: {alertDate}\n"
            instruments = (",").join(self.meta['ALERT']['event']['instruments'])
            data += f"Instruments: {instruments}\n"
            data += f"Localisation: {self.meta['HEADER']['CREATOR']}\n"

            far = 1 / (float(self.meta['ALERT']['event']['far']) * 60. * 60. * 24.)
            if far > 1000:
                far /= 365.
                data += f"FAR: 1 per {far:0.1f} yrs\n"
            else:
                data += f"FAR: 1 per {far:0.1f} days\n"

            if 'DISTMEAN' in self.meta['HEADER']:
                data += f"Dist: {self.meta['HEADER']['DISTMEAN']:.2f} (±{self.meta['HEADER']['DISTSTD']:.2f}) Mpc\n"
                ho = 67
                zlow = (ho * (self.meta['HEADER']['DISTMEAN'] - self.meta['HEADER']['DISTSTD'])) / 3e5
                zhigh = (ho * (self.meta['HEADER']['DISTMEAN'] + self.meta['HEADER']['DISTSTD'])) / 3e5
                if zlow < 0:
                    zlow = 0
                data += f"Redshift: {zlow:0.2}<z<{zhigh:0.2}\n"

            data += "\n"
            if "classification" in self.meta['ALERT']['event']:
                for k, v in self.meta['ALERT']['event']['classification'].items():
                    data += f"{k}: {v:.2f}\n"

            data += "\n"
            if "properties" in self.meta['ALERT']['event']:
                for k, v in self.meta['ALERT']['event']['properties'].items():
                    data += f"{k}: {v:.2f}\n"

            plt.text(3.42, 0.2, data, ha='left', va='top', fontsize=5, linespacing=1.8)

        # this = ax.clabel(line_c, inline=True, fontsize=6, colors=['#93a1a1'], fmt='{:.0f} '.format)

        ax.tick_params(axis='x', labelsize=12)
        ax.tick_params(axis='y', labelsize=12)
        ax.grid(color='#657b83', alpha=0.2, linestyle='dotted')
        plt.grid(True)
        plt.legend(handles=handles, loc='upper left', scatterpoints=1, bbox_to_anchor=(1.01, 1), fontsize=6)
        ax.xaxis.zorder = 40

        plt.savefig(self.outputFolder + "/skymap.png", bbox_inches='tight', dpi=300)

        plt.close()

        self.log.debug('completed the ``convert`` method')
        return None



#!/usr/bin/env python
# encoding: utf-8
"""
*Flatten a Healpix Multi-Order Map*

:Author:
    David Young

:Date Created:
    March 24, 2023
"""
from fundamentals import tools
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'


def flatten_healpix_map(
        log,
        mapPath,
        nside=64):
    """flatten a multiorder healpix map to a specific nside*

    **Key Arguments:**
        - ``log`` -- logger
        - ``mapPath`` -- path to the multiorder map
        - ``nside`` -- the nside index to flatten the map to. Default *64* (~0.9 deg2 pixels)

    **Return:**
        - ``skymap`` -- the astropy table of the map

    ```eval_rst
    .. todo::

        - create a sublime snippet for usage
        - add a tutorial about ``subtract_calibrations`` to documentation
    ```

    ```python
    from gocart.commonutils import flatten_healpix_map
    skymap = flatten_healpix_map(
        log=log,
        mapPath=pathToOutputDir + "/bayestar.multiorder.fits",
        nside=64
    )
    ```
    """
    log.debug('starting the ``flatten_healpix_map`` function')

    from mhealpy.pixelfunc.moc import uniq2range
    import tempfile
    from astropy.io import fits
    from astropy.table import Table
    import astropy_healpix as ah
    import astropy.units as u
    import numpy as np
    import pandas as pd
    from tabulate import tabulate

    # OPEN MAP TO DATAFRAME
    skymap = Table.read(mapPath)
    tableData = skymap.to_pandas()
    # FIND LEVEL AND NSIDE PIXEL INDEX FOR EACH MULTI-RES PIXEL
    tableData['LEVEL'], tableData['IPIX'] = ah.uniq_to_level_ipix(tableData['UNIQ'])
    tableData['NSIDE'] = ah.level_to_nside(tableData['LEVEL'])
    # DETERMINE THE PIXEL AREA AND PROB OF EACH PIXEL
    tableData['AREA'] = ah.nside_to_pixel_area(tableData['NSIDE']).to_value(u.steradian)
    tableData['PROB'] = tableData['AREA'] * tableData["PROBDENSITY"]

    level = ah.nside_to_level(nside)

    # UPSAMPLE TABLE
    # FIND THE HIGH-LEVEL PIXEL INDEXES FOR EACH UNIQ PIXEL
    # MAKE A NEW DATAFRAME WITH IPIX64, UNIQ
    # MATCH EACH NEW UNIQ AGAIN ORIGINAL FRAME UNIQ TO GENERATE DATA FOR IPIX^$ FRAME
    mask = tableData['NSIDE'] <= nside
    upTable = tableData.loc[mask].copy()
    upTable.reset_index(inplace=True)
    this = uniq2range(nside, upTable['UNIQ'])
    upTable[f'IPIX{nside}'] = [list(range(i, j)) for i, j in zip(this[0], this[1])]
    upTable[f'UNIQLIST'] = upTable.apply(lambda x: [x['UNIQ']] * len(x[f'IPIX{nside}']), axis=1)

    ipixNside = np.concatenate(upTable[f'IPIX{nside}'])
    uniqList = np.concatenate(upTable[f'UNIQLIST'])
    # CREATE DATA FRAME FROM A DICTIONARY OF LISTS
    myDict = {
        f'IPIX{nside}': ipixNside,
        f'UNIQ': uniqList
    }
    upTable = pd.DataFrame(myDict)
    # MERGE DATAFRAMES
    upTable = upTable.merge(tableData, on=['UNIQ'], how='inner')[[f'IPIX{nside}', 'PROBDENSITY', 'DISTMU', 'DISTSIGMA', 'DISTNORM']]
    pixArea = ah.nside_to_pixel_area(nside).to_value(u.steradian)
    upTable["PROB"] = upTable["PROBDENSITY"] * pixArea
    # REMOVE COLUMN FROM DATA FRAME
    upTable.drop(columns=['PROBDENSITY'], inplace=True)

    # DOWNSAMPLE TABLE
    mask = tableData['NSIDE'] > nside
    downTable = tableData.loc[mask].copy()
    downTable.reset_index(inplace=True)

    # FIND THE PIXEL INDEX AT ORDER NSIDE
    downTable[f'IPIX{nside}'] = np.floor_divide(tableData.loc[mask, 'IPIX'], np.power(4, (tableData.loc[mask, 'LEVEL'].values - level)))
    # GROUP RESULTS
    downTable = downTable.groupby([f'IPIX{nside}']).agg({'PROB': 'sum', 'DISTMU': 'mean', 'DISTSIGMA': 'mean', 'DISTNORM': 'mean'})
    downTable.reset_index(inplace=True)

    skymap = pd.concat([downTable, upTable], ignore_index=True)
    # SORT BY COLUMN NAME
    skymap.sort_values([f'IPIX{nside}'],
                       ascending=[True], inplace=True)

    # SET INDEX AND SORT DATA FRAME
    skymap.reset_index(inplace=True)
    skymap.drop(columns=[f'IPIX{nside}', 'index'], inplace=True)

    # REMOVE FILTERED ROWS FROM DATA FRAME
    mask = (skymap['DISTMU'].isnull())
    skymap.loc[mask, 'DISTMU'] = np.inf

    log.debug('completed the ``flatten_healpix_map`` function')
    return skymap



#!/usr/bin/env python
# encoding: utf-8
"""
*Convert Healpix Map to a Cartesian WCS*

:Author:
    David Young

:Date Created:
    March 23, 2023
"""
from fundamentals import tools
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'


class healpix2cart(object):
    """
    *Take a healpix map and convert to a rectilinear projection.*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``mapPath`` -- path the the healpix map

    **Usage:**

    To setup your logger and settings, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    To initiate a healpix2cart object, use the following:

    ```python
    from gocart.convert import healpix2cart
    converter = healpix2cart(
        log=log,
        mapPath=pathToOutputDir + "/bayestar.multiorder.fits",
        settings=settings
    )
    wcs, mapDF, header = converter.convert()
    ```

    """

    def __init__(
            self,
            log,
            mapPath,
            settings=False,

    ):
        self.log = log
        log.debug("instansiating a new 'healpix2cart' object")
        self.settings = settings
        self.mapPath = mapPath
        # xt-self-arg-tmpx

        return None

    def convert(self):
        """
        *convert the healpix map to a cartesian WCS and image*

        **Return:**
            - ``wcs`` -- an astropy wcs object
            - ``mapDF`` -- the map converted cartesian format and recorded in a pandas dataframe.
            - ``header`` -- the map header

        **Usage:**

        ```python
        from gocart.convert import healpix2cart
        converter = healpix2cart(
            log=log,
            mapPath=pathToOutputDir + "/bayestar.multiorder.fits",
            settings=settings
        )
        wcs, mapDF, header = converter.convert()
        ```
        """
        self.log.debug('starting the ``get`` method')

        import matplotlib.pyplot as plt
        from astropy.table import Table
        import astropy.units as u
        import pandas as pd
        import numpy as np
        import astropy_healpix as ah
        from astropy import wcs as awcs
        import healpy as hp

        wcs, mapDF = create_wcs_and_pixels(self.log)

        # CONVERT HEALPIX MAP TO DATAFRAME
        skymap = Table.read(self.mapPath)
        tableData = skymap.to_pandas()

        # FIND LEVEL AND NSIDE PIXEL INDEX FOR EACH MULTI-RES PIXEL
        tableData['LEVEL'], tableData['IPIX'] = ah.uniq_to_level_ipix(tableData['UNIQ'])
        tableData['NSIDE'] = ah.level_to_nside(tableData['LEVEL'])
        # DETERMINE THE PIXEL AREA AND PROB OF EACH PIXEL
        tableData['AREA'] = ah.nside_to_pixel_area(tableData['NSIDE']).to_value(u.steradian)
        tableData['PROB'] = tableData['AREA'] * tableData["PROBDENSITY"]

        # DETERMINE THE INDEX OF MULTI-RES PIX AT HIGHEST HEALPIX RESOLUTION
        max_level = 29
        max_nside = ah.level_to_nside(max_level)
        tableData['INDEX29'] = tableData['IPIX'] * (2**(max_level - tableData['LEVEL']))**2

        # DETERMINE THE HIGH-RES PIXEL LOCATION FOR EACH RA AND DEC
        match_ipix = ah.lonlat_to_healpix(mapDF["ra"].values * u.deg, mapDF["dec"].values * u.deg, max_nside, order='nested')

        # RETURNS THE INDICES THAT WOULD SORT THIS ARRAY
        sorter = np.argsort(tableData['INDEX29'])
        # FIND INDICES WHERE ELEMENTS SHOULD BE INSERTED TO MAINTAIN ORDER -- CLOSET MATCH TO THE RIGHT
        matchedIndices = sorter[np.searchsorted(tableData['INDEX29'].values, match_ipix, side='right', sorter=sorter) - 1]

        # MERGE TABLES
        mapDF = pd.concat([mapDF, tableData.iloc[matchedIndices].set_index(mapDF.index)], axis=1)

        # CLEAN UP COLUMNS
        mapDF.columns = [d.upper() for d in mapDF.columns]
        mapDF.drop(columns=['AREA', 'UNIQ', 'LEVEL', 'IPIX', 'NSIDE', 'PROB', 'INDEX29'], inplace=True)
        # ADD PIXEL PROB
        mapDF["PROBDENSITY_DEG2"] = mapDF["PROBDENSITY"] / (1 * u.steradian).to_value(u.degree * u.degree)
        mapDF["PROB"] = mapDF["PROBDENSITY_DEG2"] * mapDF["PIXEL_AREA_DEG2"]

        # SANITY CHECK
        totalProb = mapDF["PROB"].sum()
        # print(totalProb)

        self.log.debug('completed the ``get`` method')
        return wcs, mapDF, skymap.meta


def create_wcs_and_pixels(log):
    """*create the all-sky rectilinear wcs*

    **Key Arguments:**
        - ``log`` -- logger

    **Return:**
        - ``wcs`` -- the cartesian wcs.
        - ``mapDF`` -- the pixel data in a dataframe.

    **Usage:**

    ```python
    from gocart.convert import create_wcs_and_pixels
    wcs, mapDF = create_wcs_and_pixels(log=log)
    ```
    """
    log.debug('starting the ``create_wcs_and_pixels`` method')

    from astropy.wcs import WCS
    import pandas as pd
    import numpy as np
    import astropy.units as u

    # CREATE A NEW WCS OBJECT.
    wcs = WCS(naxis=2)

    # DETERMINE THE PIXEL GRID X,Y RANGES
    pixelSizeDeg = 1.
    raRange = 360
    decRange = 180
    xRange = int(raRange / pixelSizeDeg)
    yRange = int(decRange / pixelSizeDeg)

    # SET THE PIXEL SIZE
    wcs.wcs.cdelt = np.array([pixelSizeDeg, pixelSizeDeg])
    # SET THE REFERENCE PIXEL TO THE CENTRE PIXEL .. BOTTOM LEFT PIXEL IS 0,0 IN PYTHON (HENCE ADDTION OF 1)
    wcs.wcs.crpix = [(xRange + 1) / 2., (yRange + 1) / 2.]

    # FOR AN ORTHOGONAL GRID THE CRVAL2 VALUE MUST BE ZERO AND CRPIX2
    # MUST REFLECT THIS
    wcs.wcs.crpix[1] -= wcs.wcs.crval[1] / wcs.wcs.cdelt[1]
    wcs.wcs.crval[1] = 0
    wcs.wcs.crval[0] = raRange / 2

    # SET COORDINATE TYPE TO CARTESIAN
    wcs.wcs.ctype = ["RA---CAR", "DEC--CAR"]

    # CREATE THE DATA GRID
    x = np.arange(0, xRange, 1)
    y = np.arange(0, yRange, 1)
    X, Y = np.meshgrid(x, y)

    # FITS FORMAT -- BOTTOM LEFT PIXEL CENTRE IS 1,1, BUT WE ARE WORKING WITH PYTHON SO USE 0,0
    ra, dec = wcs.wcs_pix2world(X, Y, 0)
    area = np.sin(np.deg2rad(np.abs(dec + 90))) * pixelSizeDeg ** 2

    # CREATE DATA FRAME FROM A DICTIONARY OF LISTS
    myDict = {
        "pixel_x": X.ravel(),
        "pixel_y": Y.ravel(),
        "ra": ra.ravel(),
        "dec": dec.ravel(),
        "pixel_area_deg2": area.ravel()
    }
    mapDF = pd.DataFrame(myDict)

    log.debug('completed the ``create_wcs_and_pixels`` method')
    return wcs, mapDF

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/bin/env python
# encoding: utf-8
"""
*Parse Ligo-Virgo-Kagra GCN Notices*

:Author:
    David Young

:Date Created:
    March 19, 2023
"""
from gocart.commonutils import generate_skymap_stats
from gocart.convert import ascii
from fundamentals import tools
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'


class lvk(object):
    """
    *The LVK event parser*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``record`` -- the kafka record to parse.
        - ``plugins`` -- run the plugin script found in `~/.config/gocart/plugins` every time an alert is read

    **Usage:**

    To setup your logger and settings, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    To parse LVK kafka alerts, use the following:

    ```python
    from gocart.parsers import lvk
    parser = lvk(
        log=log,
        record=record,
        settings=settings,
        plugins=True
    ).parse()
    ```

    """

    def __init__(
            self,
            log,
            record,
            settings=False,
            plugins=False

    ):
        import json

        self.log = log
        log.debug("instansiating a new 'lvk' object")
        self.settings = settings
        self.record = json.loads(record)
        self.plugins = plugins

        import inspect
        import yaml
        # COLLECT ADVANCED SETTINGS IF AVAILABLE
        parentDirectory = os.path.dirname(__file__)
        advs = parentDirectory + "/advanced_settings.yaml"
        level = 0
        exists = False
        count = 1
        while not exists and len(advs) and count < 10:
            count += 1
            level -= 1
            exists = os.path.exists(advs)
            if not exists:
                advs = "/".join(parentDirectory.split("/")
                                [:level]) + "/advanced_settings.yaml"
        if not exists:
            advs = {}
        else:
            with open(advs, 'r') as stream:
                advs = yaml.safe_load(stream)
        # MERGE ADVANCED SETTINGS AND USER SETTINGS (USER SETTINGS OVERRIDE)
        self.settings = {**advs, **self.settings}

        # WHICH EVENTS ARE WE TO PARSE?
        parse_mock_events = self.settings["lvk"]["parse_mock_events"]
        parse_real_events = self.settings["lvk"]["parse_real_events"]

        if self.record['superevent_id'][0] == 'M' and not parse_mock_events:
            return
        if self.record['superevent_id'][0] != 'M' and not parse_real_events:
            return

        # WHERE TO DOWNLOAD MAPS TO
        if "download_dir" in self.settings["lvk"] and self.settings["lvk"]["download_dir"]:
            self.download_dir = self.settings["lvk"]["download_dir"]
            # MAKE RELATIVE HOME PATH ABSOLUTE
            from os.path import expanduser
            home = expanduser("~")
            if self.download_dir == "~":
                self.download_dir = self.download_dir.replace("~", home)
        else:
            self.download_dir = "."

        self.mockDir = self.download_dir + "/mockevents/"
        self.eventDir = self.download_dir + "/superevents/"
        if parse_mock_events:
            if not os.path.exists(self.mockDir):
                os.makedirs(self.mockDir)
        if parse_real_events:
            if not os.path.exists(self.eventDir):
                os.makedirs(self.eventDir)

        return None

    def parse(self):
        """
        *parse the lvk events and write meta data and maps to file
        """
        self.log.debug('starting the ``parse`` method')

        from base64 import b64decode
        from io import BytesIO
        from astropy.table import Table
        import astropy_healpix as ah
        import numpy as np
        from astropy.time import Time
        from datetime import datetime
        import yaml
        import json

        # WHICH EVENTS ARE WE TO PARSE?
        parse_mock_events = self.settings["lvk"]["parse_mock_events"]
        parse_real_events = self.settings["lvk"]["parse_real_events"]
        if self.record['superevent_id'][0] == 'M' and not parse_mock_events:
            return
        if self.record['superevent_id'][0] != 'M' and not parse_real_events:
            return

        print("\n----------------------------------------")
        if "event" in self.record and self.record["event"]:
            timeDelta = (Time(self.record["time_created"], scale='utc') - Time(self.record["event"]["time"], scale='utc')).to_value(unit='min')
            far = 1 / (float(self.record['event']['far']) * 60. * 60. * 24.)
            if far > 1000:
                far /= 365.
                far = f"1 per {far:0.1f} yrs"
            else:
                far = f"1 per {far:0.1f} days"
            print(f'EVENT: {self.record["superevent_id"]} detected at {self.record["event"]["time"].replace("Z","")} UTC ({self.record["event"]["group"]})')
            print(f'ALERT: {self.record["alert_type"].replace("_"," ")} reported at {self.record["time_created"].replace("Z","")} UTC (+{timeDelta:.2f} mins)')
            print(f'FAR: {far}')
            if "classification" in self.record['event']:
                for k, v in self.record['event']['classification'].items():
                    self.record['event']['classification'][k] = float(f'{v:.2f}')
                print(f"CLASSIFICATION: {self.record['event']['classification']}")
            if "properties" in self.record['event']:
                print(f"PROPERTIES: {self.record['event']['properties']})")
        else:
            print(f'EVENT: {self.record["superevent_id"]}')
            print(f'ALERT: {self.record["alert_type"].replace("_"," ")} reported at {self.record["time_created"].replace("Z","")} UTC')

        # PARSE SKY MAP
        header, extras, fitsPath = {}, {}, None
        localisation = False
        if self.record.get('event', {}):
            skymap_str = self.record.get('event', {}).pop('skymap')

            if skymap_str:
                # Decode, parse skymap, and print most probable sky location
                skymap_bytes = b64decode(skymap_str)
                skymap = Table.read(BytesIO(skymap_bytes))
                localisation = skymap.meta["CREATOR"].lower()
                header = {k: v for k, v in skymap.meta.items() if k != "HISTORY"}
                # GENERATE SOME EXTRA STATS
                extras = generate_skymap_stats(
                    log=self.log,
                    skymap=skymap,
                )

        # MERGE HEADER AND ALERT INTO ONE FILE
        try:
            self.record.pop('external_coinc')
        except:
            pass
        meta = {"HEADER": header, "ALERT": self.record, "EXTRA": extras}

        # DOES ALERT PASS THE FILTERS?
        if 'filters' in self.settings["lvk"]:
            if self.settings["lvk"]["filters"]:
                if not self.filter_alert(meta):
                    print("----------------------------------------\n\n")
                    return
        print("----------------------------------------\n\n")

        # DON'T WRITE RETRACTION ALERT IF NO OTHER ALERT EXISTS ON FILE
        if self.record['alert_type'].lower() == "retraction":
            if self.record["superevent_id"][0] == 'M':
                eventDir = self.mockDir + self.record["superevent_id"]
            else:
                eventDir = self.eventDir + self.record["superevent_id"]
            if not os.path.exists(eventDir):
                return

        # ADD EVENT FILTERING HERE
        # ONCE WE HAVE DECIDED TO SAVE THE EVENT/ALERT
        # RECURSIVELY CREATE MISSING DIRECTORIES
        alertTime = self.record["time_created"].replace("-", "").replace(":", "").replace(" ", "").replace("Z", "")
        if self.record["superevent_id"][0] == 'M':
            alertDir = self.mockDir + self.record["superevent_id"] + "/" + alertTime + "_" + self.record["alert_type"].lower()
        else:
            alertDir = self.eventDir + self.record["superevent_id"] + "/" + alertTime + "_" + self.record["alert_type"].lower()
        if not os.path.exists(alertDir):
            os.makedirs(alertDir)

        if self.settings["lvk"]["json"]:
            jsonName = self.record["superevent_id"] + "-" + self.record["alert_type"].lower() + ".json"
            jsonPath = alertDir + "/" + jsonName
            # DUMP JSON TO FILE
            writeFile = open(jsonPath, "w")
            json.dump(self.record, writeFile, indent=4)
            writeFile.close()

        # WRITE SKY MAP
        if localisation:
            fitsPath = f"{alertDir}/{localisation}.multiorder.fits"
            with open(fitsPath, "wb") as f:
                f.write(skymap_bytes)

        # WRITE META
        with open(alertDir + "/meta.yaml", 'w') as stream:
            yaml.dump(meta, stream, default_flow_style=False)

        if fitsPath:
            if self.settings["lvk"]["ascii_map"]["convert"]:
                c = ascii(
                    log=self.log,
                    mapPath=fitsPath,
                    nside=self.settings["lvk"]["ascii_map"]["nside"],
                    settings=self.settings
                )
                asciiContent = c.convert(outputFilepath=alertDir + "/skymap.csv")

            if self.settings["lvk"]["aitoff"]["convert"]:
                from gocart.convert import aitoff
                c = aitoff(
                    log=self.log,
                    mapPath=fitsPath,
                    outputFolder=alertDir,
                    settings=self.settings,
                    meta=meta
                )
                c.convert(
                    galacticPlane=self.settings["lvk"]["aitoff"]["galactic_plane"],
                    sunmoonContour=self.settings["lvk"]["aitoff"]["sun_moon_contour"],
                    sunmoon=self.settings["lvk"]["aitoff"]["sun_moon"])

        if self.plugins:
            from os.path import expanduser
            from subprocess import Popen, PIPE, STDOUT
            home = expanduser("~")
            # GENERATE A LIST OF FILE PATHS
            pathToDirectory = home + "/.config/gocart/plugins"
            for d in os.listdir(pathToDirectory):
                filepath = os.path.join(pathToDirectory, d)
                if os.path.isfile(filepath) and d[:3] == "gp_":
                    cmd = f"""/usr/bin/env python {filepath} {alertDir}"""
                    print(f"PLUGIN: {d}")
                    try:
                        p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
                        stdout, stderr = p.communicate()
                        print(stdout.decode('utf-8'))
                        if stderr:
                            print(f"could not execute the {d} plugin. Failed with error: {stderr.decode('utf-8')}")
                    except Exception as e:
                        self.log.error(f'could not execute the {d} plugin. Failed with error: {e}')

        self.log.debug('completed the ``parse`` method')
        return lvk

    def filter_alert(
            self,
            alert):
        """*filter the alert record with filtering criteria in the settings file and return true (pass) or false (fail)*

        **Key Arguments:**
            - ``alert`` -- the alert record

        **Return:**
            - ``passing`` -- True or False. True is alert passes one or more filter
        """
        self.log.debug('starting the ``filter_alert`` method')

        filterResults = []
        for f in self.settings["lvk"]["filters"]:

            # SOME SETUP
            try:
                f['alert_types'][:] = [g.lower() for g in f['alert_types']]
            except:
                pass

            passing = True
            message = []

            matchingCriteria = [
                'alert_types',
                "ns_lower",
                "far_upper",
                "dist_upper",
                "area90_upper",
                "hasns_lower",
                "hasremnant_lower",
                "event_dir_exits"
            ]

            if 'alert_types' in f and not alert['ALERT']['alert_type'].lower() in f['alert_types']:
                passing = False
                message.append(f"Alert type is {alert['ALERT']['alert_type'].lower()}")
            if "ns_lower" in f and 'event' in alert['ALERT'] and alert['ALERT']['event'] and 'classification' in alert['ALERT']['event'] and not alert['ALERT']['event']['classification']['BNS'] + alert['ALERT']['event']['classification']['NSBH'] >= f["ns_lower"]:
                passing = False
                message.append(f"BNS+NSBH = {alert['ALERT']['event']['classification']['BNS'] + alert['ALERT']['event']['classification']['NSBH']} (< {f['ns_lower']})")
            if "far_upper" in f and 'event' in alert['ALERT'] and alert['ALERT']['event'] and not alert['ALERT']['event']['far'] < f["far_upper"]:
                passing = False
                message.append(f"FAR = {alert['ALERT']['event']['far']} (> {f['far_upper']})")
            if "dist_upper" in f and 'DISTMEAN' in alert['HEADER'] and alert['HEADER']['DISTMEAN'] and not alert['HEADER']['DISTMEAN'] < f["dist_upper"]:
                passing = False
                message.append(f"DISTMEAN = {alert['HEADER']['DISTMEAN']} (> {f['dist_upper']})")
            if "area90_upper" in f and 'EXTRA' in alert and 'area90' in alert['EXTRA'] and not alert['EXTRA']['area90'] < f["area90_upper"]:
                passing = False
                message.append(f"area90 = {alert['EXTRA']['area90']} (> {f['area90_upper']})")
            if "hasns_lower" in f and 'event' in alert['ALERT'] and alert['ALERT']['event'] and 'properties' in alert['ALERT']['event'] and not alert['ALERT']['event']['properties']['HasNS'] >= f["hasns_lower"]:
                passing = False
                message.append(f"HasNS = {alert['ALERT']['event']['properties']['HasNS']} (< {f['hasns_lower']})")
            if "hasremnant_lower" in f and 'event' in alert['ALERT'] and alert['ALERT']['event'] and 'properties' in alert['ALERT']['event'] and not alert['ALERT']['event']['properties']['HasRemnant'] >= f["hasremnant_lower"]:
                passing = False
                message.append(f"HasRemnant = {alert['ALERT']['event']['properties']['HasRemnant']} (< {f['hasremnant_lower']})")
            if "event_dir_exits" in f:
                if self.record["superevent_id"][0] == 'M':
                    eventDir = self.mockDir + self.record["superevent_id"]
                else:
                    eventDir = self.eventDir + self.record["superevent_id"]
                if not os.path.exists(eventDir):
                    passing = False
                    message.append(f"The event has not previously passed the filtering criteria")

            for k, v in f.items():
                if k not in matchingCriteria and k != "name":
                    passing = False
                    message.append(f"Unrecognised filtering criterion '{k}'")

            filterResults.append(passing)

            if passing:
                print(f"The alert passes the {f['name']} filter")
            else:
                message = (" and ").join(message)
                print(f"The alert fails the {f['name']} filter. {message}.")

        self.log.debug('completed the ``filter_alert`` method')
        if True in filterResults:
            return True
        else:
            return False

        # use the tab-trigger below for new method
        # xt-class-method



"""
*Convert mulitorder FITS binary tables to other formats*
"""
from .aitoff import aitoff
from .healpix2cart import healpix2cart
from .healpix2cart import create_wcs_and_pixels
from .ascii import ascii
